@@ -1,8 +1,11 @@
+import logging
 import re
 import struct
 import time
 import socket, select
 import threading
+import xmltodict
+import json
 try:
     import queue as queue
 except ImportError:
@@ -11,7 +14,10 @@ import netifaces
 from collections import namedtuple
 
 from . import commands
-from .utils import ValueRange
+from .utils import ValueRange, format_nri_list
+
+BUFFER_SIZE = 64 * 1024
+_LOGGER = logging.getLogger(__name__)
 
 
 class ISCPMessage(object):
@@ -182,7 +188,7 @@ def command_to_iscp(command, arguments=None, zone=None):
             parts = [norm(c) for c in re.split(command_sep, command)]
             if len(parts) >= 3:
                 zone, command = parts[:2]
-                arguments = parts[3:]
+                arguments = parts[2:]
             elif len(parts) == 2:
                 zone = default_zone
                 command = parts[0]
@@ -205,50 +211,62 @@ def command_to_iscp(command, arguments=None, zone=None):
     # (setting tuning frequency). In some cases, we might imagine
     # providing the user an API with multiple arguments (TODO: not
     # currently supported).
-    argument = arguments[0]
+    if type(arguments) is list:
+        argument = arguments[0]
+    else:
+        argument = arguments
 
-    # 1. Consider if there is a alias, e.g. level-up for UP.
+        # 1. Consider if there is a alias, e.g. level-up for UP.
     try:
         value = commands.VALUE_MAPPINGS[group][prefix][argument]
     except KeyError:
         # 2. See if we can match a range or pattern
         for possible_arg in commands.VALUE_MAPPINGS[group][prefix]:
-            if argument.isdigit():
+            if type(argument) is int or (type(argument) is str and argument.lstrip("-").isdigit() is True):
                 if isinstance(possible_arg, ValueRange):
                     if int(argument) in possible_arg:
                         # We need to send the format "FF", hex() gives us 0xff
                         value = hex(int(argument))[2:].zfill(2).upper()
+                        if prefix == 'SWL' or prefix == 'CTL':
+                            if value == '00':
+                                value = '0' + value
+                            elif value[0] != 'X':
+                                value = '+' + value
+                            elif value[0] == 'X':
+                                if len(value) == 2:
+                                    value = '-' + '0' + value[1:]
+                                value = '-' + value[1:]
                         break
 
             # TODO: patterns not yet supported
-        else:
-            raise ValueError('"{}" is not a valid argument for command '
-                             '"{}" in zone "{}"'.format(argument, command, zone))
+            else:
+                raise ValueError('"{}" is not a valid argument for command '
+                                '"{}" in zone "{}"'.format(argument, command, zone))
 
     return '{}{}'.format(prefix, value)
 
 
-def iscp_to_command(iscp_message):
-    for zone, zone_cmds in commands.COMMANDS.items():
-        # For now, ISCP commands are always three characters, which
-        # makes this easy.
-        command, args = iscp_message[:3], iscp_message[3:]
-        if command in zone_cmds:
-            if args in zone_cmds[command]['values']:
-                return zone_cmds[command]['name'], \
-                       zone_cmds[command]['values'][args]['name']
-            else:
-                match = re.match('[+-]?[0-9a-f]+$', args, re.IGNORECASE)
-                if match:
-                    return zone_cmds[command]['name'], \
-                             int(args, 16)
+def iscp_to_command(iscp_message, with_zone=False):
+    def __iscp_to_command(iscp_message):
+        for zone, zone_cmds in commands.COMMANDS.items():
+            # For now, ISCP commands are always three characters, which
+            # makes this easy.
+            command, args = iscp_message[:3], iscp_message[3:]
+            if command in zone_cmds:
+                if args in zone_cmds[command]['values']:
+                    return zone, zone_cmds[command]['name'], \
+                           zone_cmds[command]['values'][args]['name']
                 else:
-                    return zone_cmds[command]['name'], args
-
-    else:
-        raise ValueError(
-            'Cannot convert ISCP message to command: {}'.format(iscp_message))
-
+                    match = re.match('[+-]?[0-9a-f]+$', args, re.IGNORECASE)
+                    if match:
+                        return zone, zone_cmds[command]['name'], int(args, 16)
+                    else:
+                        return zone, zone_cmds[command]['name'], args
+        else:
+            raise ValueError(
+                'Cannot convert ISCP message to command: {}'.format(iscp_message))
+    zone, ret_cmd, args = __iscp_to_command(iscp_message)
+    return (zone, ret_cmd, args) if with_zone else (ret_cmd, args)
 
 def filter_for_message(getter_func, msg):
     """Helper that calls ``getter_func`` until a matching message
@@ -261,6 +279,14 @@ def filter_for_message(getter_func, msg):
         # It seems ISCP commands are always three characters.
         if candidate and candidate[:3] == msg[:3]:
             return candidate
+        elif candidate and candidate[:3] == 'MDI' and msg[:3] == 'MGS':
+            # the MGS command for grouping multiroom audio, returns an MDI message, not MGS
+            return candidate
+
+        # exception for HDMI-CEC commands (CTV) since they don't provide any response/confirmation
+        if "CTV" in msg[:3]:
+            return msg
+        
         # The protocol docs claim that a response  should arrive
         # within *50ms or the communication has failed*. In my tests,
         # however, the interval needed to be at least 200ms before
@@ -285,10 +311,42 @@ def parse_info(data):
     ''', response.strip(), re.VERBOSE).groupdict()
     return info
 
+class MessageBuffer:
+    """A receive buffer for eISCP packets to avoid partial receipt"""
+
+    def __init__(self, buffer_size):
+        self.buffer_size = buffer_size
+        self.buffer = b''
+
+    def reset(self):
+        """Reset the buffer"""
+        self.buffer = b''
+
+    @property
+    def available(self):
+        """Bytes remaining in the buffer"""
+        return self.buffer_size - len(self.buffer)
+
+    def recv(self, data: bytes):
+        """Add received bytes to the buffer"""
+        self.buffer += data
+
+    def get_message(self):
+        """Return a message if one is available in the buffer, otherwise return None"""
+        if len(self.buffer) >= 16:
+            header = eISCPPacket.parse_header(self.buffer[:16])
+            if len(self.buffer) >= header.data_size + 16:
+                packet = ISCPMessage.parse(self.buffer[16:header.data_size + 16].decode())
+                # Remove the processed message from the remaining buffer
+                self.buffer = self.buffer[16 + header.data_size:]
+                return packet
+        return None
+
+
 class eISCP(object):
     """Implements the eISCP interface to Onkyo receivers.
 
-    This uses a blocking interface. The remote end will regularily
+    This uses a blocking interface. The remote end will regularly
     send unsolicited status updates. You need to manually call
     ``get_message`` to query those.
 
@@ -351,6 +409,8 @@ class eISCP(object):
         self.host = host
         self.port = port
         self._info = None
+        self._nri = None
+        self._message_buffer = MessageBuffer(BUFFER_SIZE)
 
         self.command_socket = None
 
@@ -397,15 +457,79 @@ class eISCP(object):
     def info(self, value):
         self._info = value
 
+    @property
+    def nri(self):
+        if self._nri:
+            return self._nri
+        return self.get_nri()
+
+    @property
+    def net_services(self):
+        data = self.nri.get('netservicelist').get('netservice')
+        return format_nri_list(data)
+
+    @property
+    def zones(self):
+        data = self.nri.get('zonelist').get('zone')
+        return format_nri_list(data)
+
+    @property
+    def controls(self):
+        data = self.nri.get('controllist').get('control')
+        return format_nri_list(data)
+
+    @property
+    def functions(self):
+        data = self.nri.get('functionlist').get('function')
+        return format_nri_list(data)
+
+    @property
+    def selectors(self):
+        data = self.nri.get('selectorlist').get('selector')
+        info = format_nri_list(data)
+        # Remove Source selector
+        if info.get("Source") is not None:
+            info.pop("Source")
+        return info
+
+    @property
+    def presets(self):
+        info = {}
+        data = self.nri.get('presetlist').get('preset')
+        for item in data:
+            if item.get("id") is not None:
+                key = item.pop("id")
+                info[key] = item
+        return info
+
+    @property
+    def tuners(self):
+        info = {}
+        data = self.nri.get('tuners').get('tuner')
+        for item in data:
+            if item.get("band") is not None:
+                key = item.pop("band")
+                info[key] = item
+        return info
+
     def _ensure_socket_connected(self):
         if self.command_socket is None:
-            self.command_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            self.command_socket.settimeout(self.CONNECT_TIMEOUT)
-            self.command_socket.connect((self.host, self.port))
-            self.command_socket.setblocking(0)
+            _LOGGER.info(f"Connecting to {self.host}:{self.port}")
+            try:
+                self.command_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                self.command_socket.settimeout(self.CONNECT_TIMEOUT)
+                self.command_socket.connect((self.host, self.port))
+                self.command_socket.setblocking(0)
+                self._message_buffer.reset()
+                _LOGGER.debug(f"Connected")
+            except OSError as error:
+                _LOGGER.error("Failed to connect", error)
+                self.disconnect()
+                raise error
 
     def disconnect(self):
         try:
+            _LOGGER.info("Closed command socket")
             self.command_socket.close()
         except:
             pass
@@ -434,17 +558,27 @@ class eISCP(object):
         """
         self._ensure_socket_connected()
 
+        # There might already be another message in the buffer
+        msg = self._message_buffer.get_message()
+        if msg:
+            return msg
+
         ready = select.select([self.command_socket], [], [], timeout or 0)
         if ready[0]:
-            header_bytes = self.command_socket.recv(16)
-            header = eISCPPacket.parse_header(header_bytes)
-            body = b''
-            while len(body) < header.data_size:
-                ready = select.select([self.command_socket], [], [], timeout or 0)
-                if not ready[0]:
+            try:
+                data = self.command_socket.recv(self._message_buffer.available)
+                self._message_buffer.recv(data)
+                if len(data) == 0:
+                    _LOGGER.error(f"Received no data on socket. Disconnecting")
+                    # We have very likely been disconnected
+                    eISCP.disconnect(self)
                     return None
-                body += self.command_socket.recv(header.data_size - len(body))
-            return ISCPMessage.parse(body.decode())
+                _LOGGER.info(f"Received {len(data)} bytes")
+                return self._message_buffer.get_message()
+            except OSError as error:
+                _LOGGER.error("Disconnected from receiver", error)
+                eISCP.disconnect(self)
+        return None
 
     def raw(self, iscp_message):
         """Send a low-level ISCP message, like ``MVL50``, and wait
@@ -489,6 +623,36 @@ class eISCP(object):
         """Turn the receiver power off."""
         return self.command('power', 'off')
 
+    def group_with(self, otherIDs=[]):
+        """Create a multiroom audio / flareconnect group with the supplied device IDs.
+        Calling this without arguments or an empty list stops the multiroom audio / flareconnect group.
+        Calling this method twice with the same arguments does not generate a response from the receiver, thus causing a timeout on the message."""
+        if otherIDs:
+            # check if the supplied deviceIDs are all strings
+            for ID in otherIDs:
+                if type(ID) != str:
+                    raise ValueError('group_with needs a list object, with each device identifier as a string')
+            # construct a MGS message with a list of the device IDs
+            message='MGS<mgs zone="1"><groupid>1</groupid><maxdelay>500</maxdelay><devices>' + \
+                '<device id="%s" zoneid="1"/>'%(self.identifier) + \
+                ''.join(['<device id="%s" zoneid="1"/>'%(ID) for ID in otherIDs]) + \
+                '</devices></mgs>'
+        else:
+            # No other devices specified. Create an empty group, which stops the multiroom audio / flareconnect
+            message='MGS<mgs zone="1"><groupid>0</groupid></mgs>'
+        return self.raw(message)
+
+    def get_nri(self):
+        """Return NRI info as dict."""
+        data = self.command("dock.receiver-information=query")[1]
+        if data:
+            data = xmltodict.parse(data, attr_prefix="")
+            data = data.get("response").get("device")
+            # Cast OrderedDict to dict
+            data = json.loads(json.dumps(data))
+            self._nri = data
+        return data
+
 
 class Receiver(eISCP):
     """Changes the behaviour of :class:`eISCP` to use a background
